#!/usr/bin/env python3
# SPDX-License-Identifier: MIT

import ctypes
from socket import AF_INET6
from ipaddress import IPv6Address

class sockaddr_in6(ctypes.Structure):
    _fields_ = (
        ('_family', ctypes.c_uint16),
        ('port', ctypes.c_uint16.__ctype_be__),
        ('flowinfo', ctypes.c_uint32.__ctype_be__),
        ('_addr', ctypes.c_ubyte * 16),
        ('scope_id', ctypes.c_uint32),
    )

    def __init__(self, **kwargs):
        super().__init__(_family=self.family)
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __repr__(self):
        return repr({
            'family': self.family,
            'port': self.port,
            'flowinfo': self.flowinfo,
            'addr': self.addr,
            'scope_id': self.scope_id,
        })

    @property
    def family(self):
        return AF_INET6

    @property
    def addr(self):
        return IPv6Address(bytes(self._addr))

    @addr.setter
    def addr(self, value):
        if not isinstance(value, IPv6Address):
            value = IPv6Address(value)

        self._addr = type(self._addr).from_buffer_copy(value.packed)

#



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT

from socket import AF_INET, AF_INET6
from ipaddress import IPv4Network, IPv6Network, ip_network
from pyroute2.netlink import nla

class allowedip(nla):
    nla_map = (
        ('WGALLOWEDIP_A_UNSPEC', 'none'),
        ('WGALLOWEDIP_A_FAMILY', 'uint16'),
        ('WGALLOWEDIP_A_IPADDR', 'cdata'),
        ('WGALLOWEDIP_A_CIDR_MASK', 'uint8'),
    )

    def network(self):
        family, ipaddr, cidr_mask = (
            self.get_attr('WGALLOWEDIP_A_FAMILY'),
            self.get_attr('WGALLOWEDIP_A_IPADDR'),
            self.get_attr('WGALLOWEDIP_A_CIDR_MASK'),
        )

        try:
            return {
                AF_INET: IPv4Network,
                AF_INET6: IPv6Network,
            }[family](ipaddr, cidr_mask)
        except:
            raise NotImplementedError

    @staticmethod
    def frob(nitz):
        self = allowedip()

        if not isinstance(nitz, (IPv4Network, IPv6Network)):
            nitz = ip_network(nitz)

        if isinstance(nitz, IPv4Network):
            self['attrs'].append(('WGALLOWEDIP_A_FAMILY', AF_INET.value))
        elif isinstance(nitz, IPv6Network):
            self['attrs'].append(('WGALLOWEDIP_A_FAMILY', AF_INET6.value))
        else:
            raise NotImplementedError

        self['attrs'].append(('WGALLOWEDIP_A_IPADDR', nitz.network_address.packed))
        self['attrs'].append(('WGALLOWEDIP_A_CIDR_MASK', nitz.prefixlen))

        return self

#



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT

from math import modf
from pyroute2.netlink import nla_base

class timespec(nla_base):
    fields = (
        ('tv_sec', 'q'),
        ('tv_nsec', 'l'),
    )

    def encode(self):
        tv_sec, tv_nsec = modf(self.value)
        self['tv_sec'] = int(tv_sec)
        self['tv_nsec'] = int(tv_nsec * 1e9)
        nla_base.encode(self)

    def decode(self):
        nla_base.decode(self)
        tv_sec = int(self['tv_sec'])
        tv_nsec = int(self['tv_nsec']) / 1e9
        self.value = tv_sec + tv_nsec

#



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT

from enum import Enum
from pyroute2.netlink import nla

class peer(nla):
    from . import key, sockaddr, timespec, allowedip

    nla_map = (
        ('WGPEER_A_UNSPEC', 'none'),
        ('WGPEER_A_PUBLIC_KEY', 'key'),
        ('WGPEER_A_PRESHARED_KEY', 'key'),
        ('WGPEER_A_FLAGS', 'uint32'),
        ('WGPEER_A_ENDPOINT', 'sockaddr'),
        ('WGPEER_A_PERSISTENT_KEEPALIVE_INTERVAL', 'uint16'),
        ('WGPEER_A_LAST_HANDSHAKE_TIME', 'timespec'),
        ('WGPEER_A_RX_BYTES', 'uint64'),
        ('WGPEER_A_TX_BYTES', 'uint64'),
        ('WGPEER_A_ALLOWEDIPS', '*allowedip'),
        ('WGPEER_A_PROTOCOL_VERSION', 'uint32'),
    )

    class flag(Enum):
        REMOVE_ME = 1 << 0
        REPLACE_ALLOWEDIPS = 1 << 1

#



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT

import ctypes
from socket import AF_INET
from ipaddress import IPv4Address

class sockaddr_in(ctypes.Structure):
    _fields_ = (
        ('_family', ctypes.c_uint16),
        ('port', ctypes.c_uint16.__ctype_be__),
        ('_addr', ctypes.c_ubyte * 4),
        ('_zero', ctypes.c_byte * 8),
    )

    def __init__(self, **kwargs):
        super().__init__(_family=self.family)
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __repr__(self):
        return repr({
            'family': self.family,
            'port': self.port,
            'addr': self.addr,
        })

    @property
    def family(self):
        return AF_INET

    @property
    def addr(self):
        return IPv4Address(bytes(self._addr))

    @addr.setter
    def addr(self, value):
        if not isinstance(value, IPv4Address):
            value = IPv4Address(value)

        self._addr = type(self._addr).from_buffer_copy(value.packed)

#



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT

from base64 import b64decode
from pyroute2.netlink import nla_base

class key(nla_base):
    fields = (
        ('value', 's'),
    )

    def encode(self):
        assert isinstance(self.value, (bytes, bytearray))
        assert 32 == len(self.value)
        self['value'] = self.value
        nla_base.encode(self)

    def decode(self):
        nla_base.decode(self)
        self.value = self['value'] if self['value'] != bytes(32) else None

    @staticmethod
    def frob(nitz):
        if not isinstance(nitz, (bytes, bytearray)):
            nitz = b64decode(nitz)

        return nitz

#



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT

from struct import unpack_from
from socket import AF_INET, AF_INET6, getaddrinfo
from urllib.parse import urlparse
from pyroute2.netlink import nla_base
from ..sockaddr_in import sockaddr_in
from ..sockaddr_in6 import sockaddr_in6

class sockaddr(nla_base):
    fields = (
        ('value', 's'),
    )

    def encode(self):
        assert isinstance(self.value, (sockaddr_in, sockaddr_in6, ))
        self['value'] = bytes(self.value)
        nla_base.encode(self)

    def decode(self):
        nla_base.decode(self)
        family, = unpack_from("H", self["value"])

        try:
            type = {
                AF_INET: sockaddr_in,
                AF_INET6: sockaddr_in6,
            }[family]
        except:
            raise NotImplementedError

        if isinstance(self['value'], bytearray):
            self.value = type.from_buffer(self['value'])
        elif isinstance(self['value'], bytes):
            self.value = type.from_buffer_copy(self['value'])

    @staticmethod
    def frob(nitz):
        if isinstance(nitz, (sockaddr_in, sockaddr_in6)):
            return nitz

        DEFAULT_PORT = 51820
        url = urlparse("//" + nitz)
        family, *meh, sockaddr = getaddrinfo(url.hostname, url.port or DEFAULT_PORT)[0]

        try:
            type, *fields = {
                AF_INET: (sockaddr_in, 'addr', 'port'),
                AF_INET6: (sockaddr_in6, 'addr', 'port', 'flowinfo', 'scope_id'),
            }[family]
        except:
            raise NotImplementedError

        return type(**dict(zip(fields, sockaddr)))

#



#!/usr/bin/env python3
from setuptools import setup

from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name = "wgnlpy",
    version = "0.0.5",
    description = ("Netlink connector to WireGuard"),
    url = "https://github.com/ArgosyLabs/wgnlpy",
    author = "Derrick Lyndon Pallas",
    author_email = "derrick@argosylabs.com",
    license = "MIT",
    packages = [ "wgnlpy", "wgnlpy/nlas" ],
    install_requires = [ "pyroute2" ],
    long_description = long_description,
    long_description_content_type = "text/markdown",
    keywords = "wireguard netlink sockaddr sockaddr_in sockaddr_in6",
    classifiers = [
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: System :: Networking",
        "Topic :: Security",
        "License :: OSI Approved :: MIT License",
    ],
    include_package_data=True,
)
#



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT

from enum import Enum
from pyroute2.netlink import genlmsg

class device(genlmsg):
    VERSION = 1

    class type(Enum):
        GET_DEVICE = 0
        SET_DEVICE = 1

    def __interface(self, interface):
        if isinstance(interface, str):
            self['attrs'].append(('WGDEVICE_A_IFNAME', interface))
        elif isinstance(interface, int):
            self['attrs'].append(('WGDEVICE_A_IFINDEX', interface))
        else:
            raise TypeError("interface must be int or string")

    @staticmethod
    def get_device(interface):
        self = device()
        self['cmd'] = self.type.GET_DEVICE.value
        self['version'] = self.VERSION
        self.__interface(interface)
        return self

    @staticmethod
    def set_device(interface):
        self = device()
        self['cmd'] = self.type.SET_DEVICE.value
        self['version'] = self.VERSION
        self.__interface(interface)
        return self

    from . import key, peer

    nla_map = (
        ('WGDEVICE_A_UNSPEC', 'none'),
        ('WGDEVICE_A_IFINDEX', 'uint32'),
        ('WGDEVICE_A_IFNAME', 'asciiz'),
        ('WGDEVICE_A_PRIVATE_KEY', 'key'),
        ('WGDEVICE_A_PUBLIC_KEY', 'key'),
        ('WGDEVICE_A_FLAGS', 'uint32'),
        ('WGDEVICE_A_LISTEN_PORT', 'uint16'),
        ('WGDEVICE_A_FWMARK', 'uint32'),
        ('WGDEVICE_A_PEERS', '*peer'),
    )

    class flag(Enum):
        REPLACE_PEERS = 1 << 0

#



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT

from pyroute2 import netlink

class WireGuard(object):
    from .nlas import device as __device

    def __init__(self, **kwargs):
        self.__socket = netlink.generic.GenericNetlinkSocket()
        self.__socket.bind('wireguard', self.__device)

    def __get(self, device):
        flags = netlink.NLM_F_ACK | netlink.NLM_F_REQUEST | netlink.NLM_F_DUMP
        return self.__socket.nlm_request(device, msg_type=self.__socket.prid, msg_flags=flags)

    def __set(self, device):
        flags = netlink.NLM_F_ACK | netlink.NLM_F_REQUEST
        return self.__socket.nlm_request(device, msg_type=self.__socket.prid, msg_flags=flags)

    def get_interface(self, interface, spill_private_key=False, spill_preshared_keys=False):
        device = self.__device.get_device(interface)
        messages = self.__get(device)

        class WireGuardInfo(object):
            def __init__(self, messages, spill_private_key, spill_preshared_keys):
                self.ifindex = messages[0].get_attr('WGDEVICE_A_IFINDEX')
                self.ifname = messages[0].get_attr('WGDEVICE_A_IFNAME')
                private_key = messages[0].get_attr('WGDEVICE_A_PRIVATE_KEY')
                if not spill_private_key:
                    private_key = private_key is not None
                self.private_key = private_key
                self.public_key = messages[0].get_attr('WGDEVICE_A_PUBLIC_KEY')
                self.listen_port = messages[0].get_attr('WGDEVICE_A_LISTEN_PORT')
                self.fwmark = messages[0].get_attr('WGDEVICE_A_FWMARK')

                self.peers = { }

                for message in messages:
                    for peer in message.get_attr('WGDEVICE_A_PEERS') or []:
                        public_key = peer.get_attr('WGPEER_A_PUBLIC_KEY')
                        assert public_key is not None
                        if public_key not in self.peers:
                            preshared_key = peer.get_attr('WGPEER_A_PRESHARED_KEY')
                            if not spill_preshared_keys:
                                preshared_key = preshared_key is not None
                            self.peers[public_key] = {
                                'preshared_key': preshared_key,
                                'endpoint': peer.get_attr('WGPEER_A_ENDPOINT'),
                                'persistent_keepalive_interval': peer.get_attr('WGPEER_A_PERSISTENT_KEEPALIVE_INTERVAL'),
                                'last_handshake_time': peer.get_attr('WGPEER_A_LAST_HANDSHAKE_TIME'),
                                'rx_bytes': peer.get_attr('WGPEER_A_RX_BYTES'),
                                'tx_bytes': peer.get_attr('WGPEER_A_TX_BYTES'),
                                'allowedips': [],
                                'protocol_version': peer.get_attr('WGPEER_A_PROTOCOL_VERSION'),
                            }
                        for allowedip in peer.get_attr('WGPEER_A_ALLOWEDIPS') or []:
                            self.peers[public_key]['allowedips'].append(allowedip.network())

            def __repr__(self):
                return repr({
                    'ifindex': self.ifindex,
                    'ifname': self.ifname,
                    'private_key': self.private_key,
                    'public_key': self.public_key,
                    'listen_port': self.listen_port,
                    'fwmark': self.fwmark,
                    'peers': self.peers,
                })

        return WireGuardInfo(messages, spill_private_key, spill_preshared_keys)

    def set_interface(self, interface,
            private_key=None,
            listen_port=None,
            fwmark=None,
            replace_peers=False,
            ):

        device = self.__device.set_device(interface)

        if replace_peers:
            device['attrs'].append(('WGDEVICE_A_FLAGS', device.flag.REPLACE_PEERS.value))

        if private_key is not None:
            device['attrs'].append(('WGDEVICE_A_PRIVATE_KEY', device.key.frob(private_key)))

        if listen_port is not None:
            device['attrs'].append(('WGDEVICE_A_LISTEN_PORT', listen_port))

        if fwmark is not None:
            device['attrs'].append(('WGDEVICE_A_FWMARK', fwmark))

        return self.__set(device)

    def remove_peers(self, interface, *public_keys):
        device = self.__device.set_device(interface)
        device['attrs'].append(('WGDEVICE_A_PEERS', []))

        for public_key in public_keys:
            peer = self.__device.peer()
            peer['attrs'].append(('WGPEER_A_PUBLIC_KEY', peer.key.frob(public_key)))
            peer['attrs'].append(('WGPEER_A_FLAGS', peer.flag.REMOVE_ME.value))
            device.get_attr('WGDEVICE_A_PEERS').append(peer)

        return self.__set(device)

    def set_peer(self, interface, public_key,
            preshared_key=None,
            endpoint=None,
            persistent_keepalive_interval=None,
            allowedips=None,
            replace_allowedips=None,
            ):

        device = self.__device.set_device(interface)

        peer = device.peer()
        peer['attrs'].append(('WGPEER_A_PUBLIC_KEY', peer.key.frob(public_key)))

        if replace_allowedips is None and allowedips is not None:
            replace_allowedips = True

        if replace_allowedips:
            peer['attrs'].append(('WGPEER_A_FLAGS', peer.flag.REPLACE_ALLOWEDIPS.value))

        if preshared_key is not None:
            peer['attrs'].append(('WGPEER_A_PRESHARED_KEY', peer.key.frob(preshared_key)))

        if endpoint is not None:
            peer['attrs'].append(('WGPEER_A_ENDPOINT', self.__device.peer.sockaddr.frob(endpoint)))

        if persistent_keepalive_interval is not None:
            peer['attrs'].append(('WGPEER_A_PERSISTENT_KEEPALIVE_INTERVAL', persistent_keepalive_interval))

        if allowedips is not None:
            peer['attrs'].append(('WGPEER_A_ALLOWEDIPS', []))

            for allowedip in allowedips:
                peer.get_attr('WGPEER_A_ALLOWEDIPS').append(self.__device.peer.allowedip.frob(allowedip))

        device['attrs'].append(('WGDEVICE_A_PEERS', [peer]))
        return self.__set(device)

    def replace_allowedips(self, interface, *public_keys):
        device = self.__device.set_device(interface)
        device['attrs'].append(('WGDEVICE_A_PEERS', []))

        for public_key in public_keys:
            peer = self.__device.peer()
            peer['attrs'].append(('WGPEER_A_PUBLIC_KEY', peer.key.frob(public_key)))
            peer['attrs'].append(('WGPEER_A_FLAGS', peer.flag.REPLACE_ALLOWEDIPS.value))
            device.get_attr('WGDEVICE_A_PEERS').append(peer)

        return self.__set(device)

#
